#include <stdio.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <ft2build.h>
#include "gifsave.h"
#include FT_FREETYPE_H

#define WIDTH		320
#define HEIGHT		160
#define MAX_FONT	3

FT_Library    	_ftlib;
FT_Face       	_face;
FT_GlyphSlot  	_slot;
FT_Vector     	_pen;
FT_Error      	_error;
int		_fontIndex = 0;
int		_fontSize  = 12;
int		_width = 0;
int		_height = 0;
int		_red = 0;
int		_green = 0;
int		_blue = 0;
char*		_renderText = "浮云逾子意";
char*		_fileName = "aaa.gif";
char*		_baseDir = "./pub/";
char*		_fontFaces [MAX_FONT] = { "hdzb_5.ttf", "hdzb_7.ttf", "hdzb_74.ttf" };
unsigned char	_image [WIDTH][HEIGHT];

/* fatal
 * print error message then exit
 */
void fatal (char* msg)
{
//	printf ("%s\n", msg);
	FT_Done_Face (_face);
	FT_Done_FreeType (_ftlib);
	exit (1);
}

/* draw_bitmap
 * copy each glyth to _image
 */
void draw_bitmap( FT_Bitmap* bitmap, FT_Int x, FT_Int y)
{
	FT_Int  i, j, p, q;
	FT_Int  xx = x + _slot->bitmap_left;
	// FIXME: cheating, don't know how to calculate the baseline, just center text vertically
	FT_Int  yy = (_height - bitmap->rows) >> 1;
	FT_Int  x_max = xx + bitmap->width;
	FT_Int  y_max = yy + bitmap->rows;
	
	for ( i = xx, p = 0; i < x_max; i++, p++ )
	{
		for ( j = yy, q = 0; j < y_max; j++, q++ )
		{
			if ( i >= WIDTH || j >= HEIGHT ) continue;
			_image[i][j] |= bitmap->buffer[q * bitmap->width + p];
		}
	}
}

/* hex2int
 * convert a two character hex string, e.g: 'FF' to int
 */
int hex2int(char* str)
{
	int code1 = tolower(str[0]);
	int code2 = tolower(str[1]);
	if (code1 >= 97) code1 -= 87; else code1 -= 48;
	if (code2 >= 97) code2 -= 87; else code2 -= 48;
	return code1 * 16 + code2;
}

/*
 * parse command line options
 * -f font -s size -t text -c color -b basedir -n filename
 */
static void parse_cmdline (int* argc, char** argv[])
{
	int option = 0;
	while (1)
	{
		option = getopt (*argc, *argv, "b:n:f:s:t:c:");
		if (option == -1) break;
		switch ( option )
		{
		case 'f': // font
			_fontIndex = atoi(optarg);
			if (_fontIndex < 0 || _fontIndex >= MAX_FONT) fatal ("Invalid font!");
			break;
		case 's': // font size
			_fontSize = atoi(optarg);
			if (_fontSize < 6 || _fontSize > 60) fatal("Invalid font size!");
			break;
		case 't': // text
			_renderText = (char*)optarg; 
			if (NULL == _renderText) fatal ("Invalid text!");
			break;
		case 'n': // filename
			_fileName = (char*)optarg; 
			if (NULL == _fileName) fatal ("Invalid filename!");
			break;
		case 'b': // filename
			_baseDir = (char*)optarg; 
			if (NULL == _baseDir) fatal ("Invalid basedir!");
			break;
		case 'c': // text color
			if (NULL == (char*)optarg) fatal ("Invalid color code!");
			_red = hex2int(optarg);
			_green = hex2int(optarg + 2);
			_blue = hex2int(optarg + 4);
		default: break;
		}
	}
}

/* get_pixel
 * get the pixel value, callback function for save_image
 */
int get_pixel (int x, int y)
{
	return (int)_image [x][y];
}

/* save image
 * save the rendered glyths to gif image
 */
int save_image (int width, int height)
{	
	int r = 255 - _red, g = 255 - _green, b = 255 - _blue;
	int i = 0;

	// create a gif with a 256 colors palette, means: 8 bits per color
	if (GIF_OK != GIF_Create(_fileName, width, height, 256, 8)) fatal ("Error creating gif file!");
	for (i = 0; i < 256; i ++)
	{
		int rr = 255 - (i * r / 255);
		int gg = 255 - (i * g / 255);
		int bb = 255 - (i * b / 255);
		// set each colors in the palette
		GIF_SetColor(i, rr, gg, bb);
	}
	// transparent color is color number 0
	GIF_SetTransparent(0);
	GIF_CompressImage (0, 0, width, height, get_pixel);
	GIF_Close();
}

/* utf8_next
 * read the next utf8 character from char* string
 */
int utf8_next (const char** pcursor, const char* end)
{
	const unsigned char*  p = (const unsigned char*)*pcursor;
	int                   ch;

	if ((const char*)p >= end) return -1;
	ch = *p++;
	if (ch >= 0x80)
	{
		int  len;
		if (ch < 0xc0)
		{
			goto BAD_DATA;
		} else if (ch < 0xe0)
		{
			len = 1;
			ch &= 0x1f;
		} else if (ch < 0xf0)
		{
			len = 2;
			ch &= 0x0f;
		} else
		{
			len = 3;
			ch &= 0x07;
		}
		while (len > 0)
		{
			if ((const char*)p >= end || ( p[0] & 0xc0 ) != 0x80) goto BAD_DATA;
			ch   = (ch << 6) | (p[0] & 0x3f);
			p   += 1;
			len -= 1;
		}
	}
	*pcursor = (const char*) p;
	return ch;
BAD_DATA: return -1;
}

/* main entry
 * parse command line options, load freetype2 lib, load font,
 * set params, render text, save to gif file, exit
 */
int main (int argc, char**  argv)
{
	int i, num;
	char filename [1024];

	parse_cmdline (&argc, &argv);
	num = strlen (_renderText);
	if (num <= 0) fatal ("Empty string!");

	// init FreeType2 library and load the font
	_error = FT_Init_FreeType (&_ftlib);
	if (_error) fatal ("Failed to init FreeType2 library!");

	memset(filename, 0, 1024 * sizeof(char));
	strcpy (filename, _baseDir);
	strcat (filename, _fontFaces [_fontIndex]);
	_error = FT_New_Face (_ftlib, filename, 0, &_face);
	if (_error) fatal ("Failed to load font file!");

	// select the character map
	_error = FT_Select_Charmap (_face, FT_ENCODING_UNICODE);
	if (_error) fatal ("Failed to select character map!");

	// set font size at 96 DPI
	_error = FT_Set_Char_Size (_face, _fontSize * 64, 0, 96, 96); 
	if (_error) fatal ("Failed to set character size!");

	_slot = _face->glyph;
	_pen.x = 0; _pen.y = 0;
	_height = _face->size->metrics.height>>6;
	const char* p = _renderText;
    	const char* pEnd = p + strlen(_renderText);
	while (1)
	{
      		int ch = utf8_next( &p, pEnd );
      		if (ch < 0) break;

		_error = FT_Load_Char (_face, ch, FT_LOAD_RENDER);
		if (_error) continue;

		draw_bitmap (&_slot->bitmap, _pen.x, _pen.y);
		_pen.x += _slot->advance.x>>6;
	}
	_width = _pen.x;
	if (_width > WIDTH || _height > HEIGHT) fatal ("Failed to render, image too big!");
	save_image(_width, _height);

	// cleanup and exit
	FT_Done_Face (_face);
	FT_Done_FreeType (_ftlib);
	return 0;
}

